from __future__ import annotations

import json
import uuid
from datetime import UTC, datetime
from pathlib import Path
from typing import Any

from mcp_eregistrations_keycloak.db.connection import get_connection
from mcp_eregistrations_keycloak.db.migrations import apply_migrations


class KcAuditLogger:
    """Lightweight SQLite audit logger for Keycloak operations."""

    def __init__(self, db_path: Path):
        self._db_path = db_path
        self._initialized = False

    @classmethod
    def for_instance(cls, instance: str | None = None) -> KcAuditLogger:
        """Create logger from instance config (mirrors KeycloakAdminClient)."""
        from mcp_eregistrations_common.config import get_instance_config
        from mcp_eregistrations_keycloak.db.connection import get_db_path

        config = get_instance_config(instance)
        keycloak_url = config.get("keycloak_url", "unknown")
        return cls(get_db_path(keycloak_url))

    async def _ensure_db(self) -> None:
        if not self._initialized:
            async with get_connection(self._db_path) as db:
                await apply_migrations(db)
            self._initialized = True

    async def record_pending(
        self,
        user_email: str | None,
        operation_type: str,
        object_type: str,
        params: dict[str, Any],
        object_id: str | None = None,
    ) -> str:
        """Record a pending operation. Returns audit_id."""
        await self._ensure_db()
        audit_id = str(uuid.uuid4())
        now = datetime.now(UTC).isoformat()
        async with get_connection(self._db_path) as db:
            await db.execute(
                """INSERT INTO audit_entries
                   (id, timestamp, user_email, operation_type, object_type,
                    object_id, params, status)
                   VALUES (?, ?, ?, ?, ?, ?, ?, 'pending')""",
                (
                    audit_id,
                    now,
                    user_email,
                    operation_type,
                    object_type,
                    object_id,
                    json.dumps(params),
                ),
            )
            await db.commit()
        return audit_id

    async def mark_success(self, audit_id: str, result: dict[str, Any]) -> None:
        """Mark an audit entry as successful."""
        async with get_connection(self._db_path) as db:
            await db.execute(
                "UPDATE audit_entries SET status='success',"
                " result=? WHERE id=? AND status='pending'",
                (json.dumps(result), audit_id),
            )
            await db.commit()

    async def mark_failed(self, audit_id: str, error_message: str) -> None:
        """Mark an audit entry as failed."""
        async with get_connection(self._db_path) as db:
            await db.execute(
                "UPDATE audit_entries SET status='failed',"
                " error_message=? WHERE id=? AND status='pending'",
                (error_message, audit_id),
            )
            await db.commit()

    async def save_rollback_state(
        self,
        audit_id: str,
        object_type: str,
        object_id: str | None,
        previous_state: dict[str, Any],
    ) -> str:
        """Save previous state for potential rollback. Returns rollback state id."""
        state_id = str(uuid.uuid4())
        now = datetime.now(UTC).isoformat()
        async with get_connection(self._db_path) as db:
            await db.execute(
                """INSERT INTO rollback_states
                   (id, audit_id, object_type, object_id, previous_state, created_at)
                   VALUES (?, ?, ?, ?, ?, ?)""",
                (
                    state_id,
                    audit_id,
                    object_type,
                    object_id,
                    json.dumps(previous_state),
                    now,
                ),
            )
            await db.commit()
        return state_id

    async def list_entries(
        self,
        limit: int = 20,
        offset: int = 0,
        object_type: str | None = None,
    ) -> list[dict[str, Any]]:
        """List audit entries with optional filtering."""
        await self._ensure_db()
        async with get_connection(self._db_path) as db:
            if object_type:
                cursor = await db.execute(
                    """SELECT * FROM audit_entries
                       WHERE object_type=?
                       ORDER BY timestamp DESC LIMIT ? OFFSET ?""",
                    (object_type, limit, offset),
                )
            else:
                cursor = await db.execute(
                    "SELECT * FROM audit_entries"
                    " ORDER BY timestamp DESC LIMIT ? OFFSET ?",
                    (limit, offset),
                )
            rows = await cursor.fetchall()
            return [
                {
                    "id": row["id"],
                    "timestamp": row["timestamp"],
                    "user_email": row["user_email"],
                    "operation_type": row["operation_type"],
                    "object_type": row["object_type"],
                    "object_id": row["object_id"],
                    "status": row["status"],
                    "params": json.loads(row["params"]) if row["params"] else None,
                    "result": json.loads(row["result"]) if row["result"] else None,
                    "error_message": row["error_message"],
                    "rolled_back_at": row["rolled_back_at"],
                }
                for row in rows
            ]

    async def get_rollback_state(self, audit_id: str) -> dict[str, Any] | None:
        """Get rollback state for an audit entry."""
        async with get_connection(self._db_path) as db:
            cursor = await db.execute(
                "SELECT * FROM rollback_states WHERE audit_id=?",
                (audit_id,),
            )
            row = await cursor.fetchone()
            if not row:
                return None
            return {
                "id": row["id"],
                "audit_id": row["audit_id"],
                "object_type": row["object_type"],
                "object_id": row["object_id"],
                "previous_state": json.loads(row["previous_state"]),
                "created_at": row["created_at"],
            }

    async def mark_rolled_back(self, audit_id: str) -> None:
        """Mark an audit entry as rolled back."""
        now = datetime.now(UTC).isoformat()
        async with get_connection(self._db_path) as db:
            await db.execute(
                "UPDATE audit_entries SET rolled_back_at=? WHERE id=?",
                (now, audit_id),
            )
            await db.commit()

    async def cleanup_old_entries(self, days: int = 30) -> int:
        """Delete rollback states older than specified days. Returns count deleted."""
        await self._ensure_db()
        async with get_connection(self._db_path) as db:
            cursor = await db.execute(
                """DELETE FROM rollback_states
                   WHERE created_at < datetime('now', ?)""",
                (f"-{days} days",),
            )
            count: int = cursor.rowcount
            await db.commit()
            return count



"""GDB service tools — list, get, create, update, set_visible, delete."""

from __future__ import annotations

from typing import Any

from mcp.types import ToolAnnotations

from mcp_eregistrations_gdb.gdb_client import GDBClient, GDBClientError, translate_error
from mcp_eregistrations_gdb.server import mcp
from mcp_eregistrations_gdb.tools import _add_optional


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def gdb_service_list(instance: str | None = None) -> dict[str, Any]:
    """List all GDB services.

    Args:
        instance: Instance name (default: auto-select).

    Returns:
        dict with count, services.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            data = await client.get("/services")
            if isinstance(data, list):
                services = data
            else:
                services = data.get("results", [])
            return {"count": len(services), "services": services}
    except GDBClientError as e:
        raise translate_error(e, resource_type="service") from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def gdb_service_get(
    service_id: int,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get a GDB service by ID.

    Args:
        service_id: Integer ID of the service.
        instance: Instance name (default: auto-select).

    Returns:
        dict with service details.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            result: dict[str, Any] = await client.get(f"/api/v1/services/{service_id}")
            return result
    except GDBClientError as e:
        raise translate_error(e, resource_type="service", resource_id=service_id) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=False))
async def gdb_service_create(
    name: str,
    path: str,
    catalog_id: int,
    template: str,
    parameters: dict[str, Any] | None = None,
    method: str | None = None,
    response_schema: dict[str, Any] | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Create a GDB service. Audited write operation.

    Args:
        name: Service name.
        path: Service URL path.
        catalog_id: ID of the parent catalog.
        template: Template identifier.
        parameters: Optional JSON parameters (default: None).
        method: HTTP method override (default: None).
        response_schema: Response JSON schema (default: None).
        instance: Instance name (default: auto-select).

    Returns:
        dict with created service details.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            body: dict[str, Any] = {
                "name": name,
                "path": path,
                "catalog_id": catalog_id,
                "template": template,
            }
            _add_optional(body, "parameters", parameters)
            _add_optional(body, "method", method)
            _add_optional(body, "responseSchema", response_schema)
            result: dict[str, Any] = await client.post(
                "/api/v1/service/modify", data=body
            )
            return result
    except GDBClientError as e:
        raise translate_error(e, resource_type="service") from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=False))
async def gdb_service_update(
    service_id: int,
    name: str | None = None,
    path: str | None = None,
    catalog_id: int | None = None,
    template: str | None = None,
    parameters: dict[str, Any] | None = None,
    method: str | None = None,
    response_schema: dict[str, Any] | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Update a GDB service. Audited write operation.

    Args:
        service_id: Integer ID of the service to update.
        name: New service name (default: None).
        path: New service URL path (default: None).
        catalog_id: New parent catalog ID (default: None).
        template: New template identifier (default: None).
        parameters: New JSON parameters (default: None).
        method: New HTTP method override (default: None).
        response_schema: New response JSON schema (default: None).
        instance: Instance name (default: auto-select).

    Returns:
        dict with updated service details.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            body: dict[str, Any] = {"id": service_id}
            _add_optional(body, "name", name)
            _add_optional(body, "path", path)
            _add_optional(body, "catalog_id", catalog_id)
            _add_optional(body, "template", template)
            _add_optional(body, "parameters", parameters)
            _add_optional(body, "method", method)
            _add_optional(body, "responseSchema", response_schema)
            result: dict[str, Any] = await client.post(
                "/api/v1/service/modify", data=body
            )
            return result
    except GDBClientError as e:
        raise translate_error(e, resource_type="service", resource_id=service_id) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=False))
async def gdb_service_set_visible(
    service_id: int,
    visible: bool,
    instance: str | None = None,
) -> dict[str, Any]:
    """Set GDB service visibility. Audited write operation.

    Args:
        service_id: Integer ID of the service.
        visible: Whether the service should be visible.
        instance: Instance name (default: auto-select).

    Returns:
        dict with service_id, visible.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            await client.put(
                f"/api/v1/service/{service_id}/visible",
                data={"value": visible},
            )
            return {"service_id": service_id, "visible": visible}
    except GDBClientError as e:
        raise translate_error(e, resource_type="service", resource_id=service_id) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=False))
async def gdb_service_delete(
    service_id: int,
    instance: str | None = None,
) -> dict[str, Any]:
    """Delete a GDB service. Audited write operation.

    Args:
        service_id: Integer ID of the service to delete.
        instance: Instance name (default: auto-select).

    Returns:
        dict with deleted (bool), service_id.
    """
    try:
        async with GDBClient.for_instance(instance) as client:
            await client.delete(f"/api/v1/service/{service_id}")
            return {"deleted": True, "service_id": service_id}
    except GDBClientError as e:
        raise translate_error(e, resource_type="service", resource_id=service_id) from e



from typing import Any

from fastmcp.exceptions import ToolError
from mcp.types import ToolAnnotations

from mcp_eregistrations_keycloak.audit.logger import KcAuditLogger
from mcp_eregistrations_keycloak.audit.models import KcObjectType
from mcp_eregistrations_keycloak.keycloak_client import (
    KeycloakAdminClient,
    KeycloakClientError,
    translate_error,
)
from mcp_eregistrations_keycloak.server import mcp
from mcp_eregistrations_keycloak.tools.validation import _resolve_realm


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_user_profile_config(
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get user profile configuration (attribute definitions, validations).

    Args:
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user profile configuration.
    """
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            result: dict[str, Any] = await client.get(f"/realms/{r}/users/profile")
            return result
    except KeycloakClientError as e:
        raise translate_error(e, resource_type="user profile config") from e


@mcp.tool()
async def kc_update_user_profile_config(
    config: dict[str, Any],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Replace the full user profile configuration. Audited write operation.

    Args:
        config: Complete user profile configuration to set.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with realm, status.
    """
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None, "update", KcObjectType.USER_PROFILE, {"realm": r}
            )
            previous = await client.get(f"/realms/{r}/users/profile")
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_PROFILE, r, previous
            )
            await client.put(f"/realms/{r}/users/profile", data=config)
            outcome = {
                "realm": r,
                "status": "updated",
                "attribute_count": len(config.get("attributes", [])),
                "previous_attribute_count": len(previous.get("attributes", [])),
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(e, resource_type="user profile config") from e


@mcp.tool()
async def kc_add_user_attribute(
    attribute: dict[str, Any],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Add a new attribute to user profile config. Audited write operation.

    Args:
        attribute: Attribute definition (must include 'name').
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with realm, attribute_name, status.
    """
    attr_name = attribute.get("name")
    if not attr_name:
        raise ToolError("Attribute must include a 'name' field.")
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "create",
                KcObjectType.USER_PROFILE,
                {"realm": r, "attribute_name": attr_name},
            )
            config = await client.get(f"/realms/{r}/users/profile")
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_PROFILE, r, config
            )
            attrs = config.get("attributes", [])
            existing_names = {a["name"] for a in attrs}
            if attr_name in existing_names:
                raise ToolError(
                    f"Attribute '{attr_name}' already exists. "
                    f"Use kc_update_user_attribute to modify it."
                )
            attrs.append(attribute)
            config["attributes"] = attrs
            await client.put(f"/realms/{r}/users/profile", data=config)
            outcome = {
                "realm": r,
                "attribute_name": attr_name,
                "status": "added",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="user attribute",
            resource_id=attr_name,
        ) from e


@mcp.tool()
async def kc_update_user_attribute(
    attribute_name: str,
    updates: dict[str, Any],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Update an existing attribute in user profile config. Audited write operation.

    Args:
        attribute_name: Name of attribute to update.
        updates: Fields to merge into the attribute definition.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with realm, attribute_name, status.
    """
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "update",
                KcObjectType.USER_PROFILE,
                {"realm": r, "attribute_name": attribute_name},
            )
            config = await client.get(f"/realms/{r}/users/profile")
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_PROFILE, r, config
            )
            attrs = config.get("attributes", [])
            found = False
            for attr in attrs:
                if attr["name"] == attribute_name:
                    attr.update(updates)
                    attr["name"] = attribute_name  # prevent name overwrite
                    found = True
                    break
            if not found:
                raise ToolError(
                    f"Attribute '{attribute_name}' not found. "
                    f"Use kc_add_user_attribute to create it."
                )
            config["attributes"] = attrs
            await client.put(f"/realms/{r}/users/profile", data=config)
            outcome = {
                "realm": r,
                "attribute_name": attribute_name,
                "status": "updated",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="user attribute",
            resource_id=attribute_name,
        ) from e


@mcp.tool()
async def kc_remove_user_attribute(
    attribute_name: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Remove an attribute from user profile config. Audited write operation.

    Args:
        attribute_name: Name of attribute to remove.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with realm, attribute_name, status.
    """
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "delete",
                KcObjectType.USER_PROFILE,
                {"realm": r, "attribute_name": attribute_name},
            )
            config = await client.get(f"/realms/{r}/users/profile")
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_PROFILE, r, config
            )
            attrs = config.get("attributes", [])
            new_attrs = [a for a in attrs if a["name"] != attribute_name]
            if len(new_attrs) == len(attrs):
                raise ToolError(
                    f"Attribute '{attribute_name}' not found in user profile config."
                )
            config["attributes"] = new_attrs
            await client.put(f"/realms/{r}/users/profile", data=config)
            outcome = {
                "realm": r,
                "attribute_name": attribute_name,
                "status": "removed",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="user attribute",
            resource_id=attribute_name,
        ) from e
